
#include "usb910conf.h"

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>

#include "usbcdc.h"
#include "usb910.h"

#ifndef ADCD
#define ADCD ADC
#endif

#define AUTO_RANGE_SUPPORT

#define poll    usbcdc_poll

#define bit_set(port,bit)	{ port |= (1<<(bit));}
#define bit_clear(port,bit)	{ port &= ~(1<<(bit));}
#define min(a,b)		((a)<(b) ? (a):(b))


#ifdef ADC_EXTENSION

#define ADC_MAX_CONF	8
static uint8_t adc_conf[ADC_MAX_CONF];
/* result format 
 *
 *  int16_t: unit 0.15625 mV (10/64 mV)
 *
 *   +-+-+----------+---+
 *   |1|1|  10      | 4 |
 *   +-+-+----------+---+
 *       |          |
 *
 */

#if defined(__AVR_ATmega88__)

/* 
 *   VCC     --  ADC3 x k (about 5/2)
 *   CURRENT --  DELTA ADC0(+) ADC1(-)
 *   AIN1    --  ADC4
 *   AIN2    --  ADC5
 *   AIN3    --  ADC6
 *   AIN4    --  ADC9
 */

#define REF_VCC		(1<<REFS0)
#define REF_V110	(3<<REFS0) 
#define REF_EXT		(0<<REFS0)
#define CHAN_ADC0	(0x00 | (1<<ADLAR))
#define CHAN_ADC1	(0x01 | (1<<ADLAR))
#define CHAN_ADC2	(0x02 | (1<<ADLAR))
#define CHAN_ADC3	(0x03 | (1<<ADLAR))
#define CHAN_ADC4	(0x04 | (1<<ADLAR))
#define CHAN_ADC5	(0x05 | (1<<ADLAR))
#define CHAN_VCC	CHAN_ADC4

const static uint8_t adc_chan_se[6] = 
{
	CHAN_ADC0, CHAN_ADC1, CHAN_ADC2, CHAN_ADC3,
	CHAN_ADC4, CHAN_ADC5
};

#else
#error "ADC_EXTENSION is not supported\n"
#endif

static inline uint8_t adc_is_busy() {
	return bit_is_set(ADCSRA,ADSC);
}

#define ALWAYS_ADEN

static inline uint16_t do_adc_native(uint8_t admux) {
	ADMUX = admux;
	uint16_t r1,r2,r3,tmp;
#ifdef ALWAYS_ADEN
	ADCSRA |= (1<<ADSC);
	while (adc_is_busy()) {
		poll();
	}
	r1 = ADCD;
	ADCSRA |= (1<<ADSC);
	while (adc_is_busy()) {
		poll();
	}
	r2 = ADCD;
	ADCSRA |= (1<<ADSC);
	while (adc_is_busy()) {
		poll();
	}
	r3 = ADCD;
	if (r2 > r1) {
		tmp = r1; r1 = r2; r2 =tmp;
	}
	if (r3 > r1) {
		tmp = r1; r1 = r3; r3 =tmp;
	}
	if (r3 > r2) {
		tmp = r2; r2 = r3; r3 =tmp;
	}
	return r2;
#else
	ADCSRA |= (1<<ADEN);
	ADCSRA |= (1<<ADSC);
	while (adc_is_busy()) {
		poll();
	}
	/* discard 1st result */
	ADCSRA |= (1<<ADSC);
	while (adc_is_busy()) {
		poll();
	}
	ADCSRA &= ~(1<<ADEN);
	return ADCD;
#endif
}

void init_adc() {
	/* setup ADC @ 1/64 clock */
#ifdef ALWAYS_ADEN
	ADCSRA = (6<<ADPS0) | (1<<ADEN);
#else
	ADCSRA = (6<<ADPS0);
#endif
	/* digial input disable */
	DDRC &=  ~( (1<<PC0)|(1<<PC1)|(1<<PC2)|(1<<PC3)|(1<<PC4)|(1<<PC5));
	PORTC &=  ~( (1<<PC0)|(1<<PC1)|(1<<PC2)|(1<<PC3)|(1<<PC4)|(1<<PC5));
	DIDR0 = (1<<ADC0D) | (1<<ADC1D) | (1<<ADC2D)
		| (1<<ADC3D) |(1<<ADC4D) | (1<<ADC5D);
}

int8_t adc_send(uint8_t size) {
	uint8_t i;
	if (size > ADC_MAX_CONF) size = ADC_MAX_CONF;
	for (i=0; i<size; i++) {
		adc_conf[i] = ext_buf[i];
	}
	return i;
}


static uint16_t vcc_val() {
	uint16_t vcc_calib;
	uint16_t result;
	uint32_t v32;

	result = do_adc_native(REF_EXT | CHAN_VCC);
	// range must be (1/8 to 1/2 )
	if ((result < (128<<6)) || (result >= (512<<6))) {
		return 0x7fff;
	} else {
		v32 = 0x40000000;
		vcc_calib = v32 / result; 
	}
	v32 = do_adc_native(REF_EXT | CHAN_VCC);
	v32 = v32 * vcc_calib;
	result = (uint16_t)(v32 >> 16);
	return result;
}

uint16_t do_adc_se(uint8_t chan);

#define SEL_RANGE_V256	(uint16_t)( 0x4000 * 0.93)

uint16_t do_adc_se(uint8_t chan) {
	uint8_t range;
	uint8_t auto_mode = 0;
	uint8_t admux;
	uint16_t result;
	range = chan & ADC_RANGE_MASK;
	if (range == ADC_RANGE_AUTO) {
		auto_mode = 1;
		range = ADC_RANGE_V512;
	}
	chan &= ADC_CHAN_MASK;
	if (chan == ADC_CHAN_VCC) {
		return vcc_val();
	}
	if (chan == ADC_CHAN_TEMP) {
		return 0;
	}
	if (chan == ADC_CHAN_AIN6) {
		return 0;
	}
	admux = adc_chan_se[chan];
#ifdef AUTO_RANGE_SUPPORT
range_down:
#endif
	if (range == ADC_RANGE_V512) {
	} else if (range == ADC_RANGE_V256) {
		admux |= (2<<REFS0);
	}
	result = do_adc_native(admux);
	if (range == ADC_RANGE_V512) {
		uint32_t v32;
		v32 = result;
		v32 = v32 * vcc_val();
		result = v32 >> 16;
#ifdef AUTO_RANGE_SUPPORT
		if (auto_mode) {
			if (result < SEL_RANGE_V256) {
				range = ADC_RANGE_V256;
				goto range_down;
			}
		}
#endif
	} else if (range == ADC_RANGE_V256) {
		result = result >> 2;
	}
	return result;
}

int8_t adc_recv(uint8_t size) {
	uint8_t i;
	uint16_t *p = (uint16_t *)ext_buf;
	size >>= 1;

	for (i=0; i<size; i++,p++) {
	    if (adc_conf[i] & ADC_CHAN_DELTA) {
		    *p = 0;
	    } else {
		    *p = do_adc_se(adc_conf[i]);
            }
	}
	return (i<<1);
}

#endif
