/*
 * Copyright (c) 2007 Koji Suzuki
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <time.h>
#include <ctype.h>
#include <unistd.h>


#if defined(_WIN32)
#define WIN32NATIVE
#endif
#if defined(WIN32NATIVE)
#include <windows.h>
#include <io.h>
#include <mmsystem.h>

static inline double dtime() {
	return (double)timeGetTime()/1.0e3;
}

void usleep(int t) {
	Sleep(1 + (t+999)/1000);
}


int set_serial(int fd) {
	HANDLE handle;
	DCB dcb;
	COMMTIMEOUTS cto;

	handle = (HANDLE)(void *)_get_osfhandle(fd);
	if (handle == INVALID_HANDLE_VALUE) {
		printf("get_osfhandle error\n");
		fflush(stdout);
		exit(1);
		//return -1;
	}
	if (!SetupComm(handle, 256, 256)) {
		printf("SetupComm error\n");
		fflush(stdout);
		return -1;
	}
	memset(&dcb, 0,  sizeof(DCB));
	dcb.DCBlength = sizeof(DCB);
	dcb.BaudRate = CBR_19200;
	dcb.fBinary = 1;
	dcb.fDtrControl = DTR_CONTROL_DISABLE;
	dcb.fRtsControl = RTS_CONTROL_DISABLE;
	dcb.ByteSize = 8;
	dcb.Parity = NOPARITY;
	dcb.StopBits = ONESTOPBIT;
	if (!SetCommState(handle, &dcb)) {
		return -1;
	}
	memset(&cto, 0, sizeof(COMMTIMEOUTS));
	cto.ReadIntervalTimeout = 500;        // ms
	cto.ReadTotalTimeoutMultiplier = 500; // ms
	cto.ReadTotalTimeoutConstant = 500;   // ms
	SetCommTimeouts(handle, &cto);
	return 0;
}

#else
#include <termios.h>
#include <sys/time.h>
#include <sys/types.h>

static inline double dtime() {
	struct timeval tv;
	double t;

	gettimeofday(&tv,0);
	return (double)(unsigned long)tv.tv_sec + (double)tv.tv_usec/1.0e6;
}

int set_serial(int fd) {
	int speed;
	struct termios term;

	tcgetattr(fd, &term);
	speed = B19200;
	cfsetispeed(&term,speed);
	cfsetospeed(&term,speed);
	cfmakeraw(&term);
	usleep(100000); /* 100 ms */
	tcsetattr(fd, TCSANOW, &term);
	/* setup done */
	return 0;
}

#ifndef O_BINARY
#define O_BINARY	0
#endif

#endif



#define DEV_CHAN	0x10
#define uint8_t unsigned char
#define int16_t short
#define uint16_t unsigned short

#define ADC_CHAN_AIN1           0x0 // using ADC4
#define ADC_CHAN_AIN2           0x1 // using ADC5
#define ADC_CHAN_AIN3           0x2 // using ADC6
#define ADC_CHAN_AIN4           0x3 // using ADC9
#define ADC_CHAN_AIN5           0x4 // internal: using ADC0
#define ADC_CHAN_AIN6           0x5 // internal: using ADC1
#define ADC_CHAN_VCC            0x6 // internal: using ADC3
#define ADC_CHAN_TEMP           0x7

#define ADC_CHAN_CURRENT        0x8 //internal
#define ADC_CHAN_AIN12          0x9
#define ADC_CHAN_AIN23          0xa
#define ADC_CHAN_AIN31          0xb
#define ADC_CHAN_AIN65          0xc //internal reverse input
#define ADC_CHAN_AIN21          0xd // reverse input
#define ADC_CHAN_AIN32          0xe // reverse input
#define ADC_CHAN_AIN13          0xf // reverse input

#define ADC_CHAN_DELTA          0x08
#define ADC_CHAN_MASK           0x0f

#define ADC_RANGE_AUTO          0x00 // Auto Range
#define ADC_RANGE_V512          0x10 // Reference VCC (+ calibrate VCC)
#define ADC_RANGE_V256          0x20 // Reference 2.56V
#define ADC_RANGE_V110          0x30 // Reference 1.10V
#define ADC_RANGE_V032          0x50 // Reference 2.56V + Gain x8
#define ADC_RANGE_V008          0x70 // Reference 2.56V + Gain x32
#define ADC_RANGE_MASK          0x70
#define ADC_BIPOLAR             0x80 // use BIN

#define ADC_MAX_CONF            4

void send_and_recv(int fd, uint8_t *buf, int slen, int rlen) {
	int r,i;
	if (slen != 0) {
	    r = write(fd, buf, slen);
	    if (r != slen ) {
		fprintf(stderr,"cmd send error r = %d %02x\n", r , buf[0]);
		exit(1);
	    }
	}
	i = 0;
	if (rlen != 0) {
re_read:
	    r = read(fd, buf+i , rlen-i);
	    if (r <= 0) {
		fprintf(stderr,"reply error r = %d %02x\n", r , buf[0]);
		exit(1);
	    }
	    if (r != rlen-i) {
		i += r;
		goto re_read;
	    }
	}
}

init_usb910(int fd) {
	int r,i;
	uint8_t buf[1024];

	for (i=0; i<10; i++) {
		r = write(fd, "S", 1);
		if (r < 1) {
			perror("write 1");
			exit(1);
		}
		usleep(1000); /* 1 ms */
		r = read(fd, buf, 7);
		if (r == 7 && (!strncmp(buf,"USB910",6)
				|| !strncmp(buf,"SERJTAG",7))) {
			buf[7] = 0;
			printf("nego OK [%s]\n", buf);
			fflush(stdout);
			r = write(fd, "a", 1);
			usleep(1000); /* 1 ms */
			r = read(fd, buf, 1);
			if (r == 1 && buf[0] == 'Y') {
				printf("auto address incliment OK\n");
				fflush(stdout);
				goto next;
			} else {
				printf("auto address incliment NG\n");
				fflush(stdout);
			}
		}
		printf("read len = %d\n",r);
		fflush(stdout);
	}
	return -1;
next:
	buf[0] = 'e';
	send_and_recv(fd, buf, 1, 1);
	if (buf[0] != 'Y') {
		fprintf(stderr,"enter extended mode Error\n",buf[0]);
	}
}

static int send_data;

void test_send(int fd2, int count) {
	int r,i;
        uint8_t buf[1024];

	send_data++;
	buf[0] = 's';	// send
	buf[1] = 9;	// test
	buf[2] = count;	// send len
	for (i=0; i< count; i++) {
		buf[3+i] = send_data;
	}
	send_and_recv(fd2, buf, 3 + count, 1);
	if (buf[0] != count) {
		printf("send count error %d %d %d\n",count,buf[0],send_data);
		exit(1);
	}
}

void test_recv(int fd2, int count, int chk) {
	int r,i,sum;
        uint8_t buf[1024];

	buf[0] = 'r';	// recv
	buf[1] = 9;	// test
	buf[2] = count;	// result len
	send_and_recv(fd2, buf, 3, 1);
	if (buf[0] != count) {
		printf("recv count error %d %d %d\n",count,buf[0],send_data);
		exit(1);
	}
	send_and_recv(fd2, buf+1, 0, count);
	if (chk) {
		if (count > chk) count = chk;
		for (i=0; i< count; i++) {
			if (buf[1+i] != (send_data & 0xff)) {
				fprintf(stderr,"read_error at %d\n",i);
				exit(1);
			}
		}
	}
}

#if defined(WIN32NATIVE)
char *port_name = "COM4";
#else
char *port_name = "/dev/ttyACM0";
#endif

main(int argc,char *argv[]) {
	int fd,c,i;
	int send_size,recv_size,loop_cnt;
	double st,et;
	uint8_t buf[1024];

	send_size = 4;
	recv_size = 4;
	loop_cnt = 1000;
	while ((c=getopt(argc,argv,"P:s:r:l:")) != EOF) {
	    switch(c) {
		case 'P':
			port_name = optarg;
			break;
		case 'l':
			loop_cnt = atoi(optarg);
			break;
		case 's':
			send_size = atoi(optarg);
			break;
		case 'r':
			recv_size = atoi(optarg);
			break;
	    }
	}
	fd = open(port_name, O_RDWR|O_BINARY);
	if (fd < 0) {
		printf("can't open serial\n");
		exit(1);
	}
	set_serial(fd);
	init_usb910(fd);

	printf("start --  send %d recv %d loops %d\n",send_size,recv_size,loop_cnt);
	fflush(stdout);
	st = dtime();
	for (i=0; i< loop_cnt; i++) {
		if (send_size > 0) {
			test_send(fd, send_size);
		}
		if (recv_size > 0) {
			test_recv(fd, recv_size,send_size);
		}
	}
	et = dtime();
	printf("end test  %5.3f sec\n", et - st);
}
